
/* ========================================================================== */

/*
 *                          S p a r s e T o o l s
 *
 *   S p a r s e   m a t r i x   e x t e n s i o n s   t o   P y S p a r s e
 */

/* ========================================================================== */

#include "Python.h"                /* Main Python header file */

#include "arrayflavor.h"

#include "pysparse/spmatrix.h"
#include "pysparse/spmatrix_api.h"
#include "pysparse/ll_mat.h"
#include "pysparse/csr_mat.h"

#define ZERO (double)0.0

/* ========================================================================== */

void init_sparsetools( void );

static PyObject *LLMat_To_Coord(  PyObject *self, PyObject *args );
static PyObject *CSRMat_To_Coord( PyObject *self, PyObject *args );
static PyObject *CSRMat_usolve(   PyObject *self, PyObject *args );
static PyObject *CSRMat_utsolve(  PyObject *self, PyObject *args );
static PyObject *CSRMat_lsolve(   PyObject *self, PyObject *args );
static PyObject *CSRMat_ltsolve(  PyObject *self, PyObject *args );

/* ========================================================================== */

/*
 *                    M o d u l e   f u n c t i o n s
 */

/* ========================================================================== */

static char LLMat_To_Coord_Doc[] = "Output coordinates of LL matrix.";

static PyObject *LLMat_To_Coord( PyObject *self, PyObject *args ) {

    /* Convert an LL matrix into coordinate format */

    PyObject      *mat;                   /* Input matrix */
    LLMatObject   *llmat;                 /* Input matrix in LL format   */
    PyArrayObject *a_row, *a_col, *a_val; /* Matrix in coordinate format */
    int            dmat[1];               /* Dimension descriptor */
    int           *pi, *pj;           /* Intermediate pointers to matrix data */
    double        *pv;
    int            i, k, elem;

    /* Input matrix is the only argument */
    if( !PyArg_ParseTuple( args, "O", &mat ) )
    return NULL;

    llmat   = (LLMatObject *)mat;
    dmat[0] = llmat->nnz;

    /* Allocate numarrays */
    a_row = (PyArrayObject *)PyArray_SimpleNew( 1, dmat, NPY_INT32 );
    a_col = (PyArrayObject *)PyArray_SimpleNew( 1, dmat, NPY_INT32 );
    a_val = (PyArrayObject *)PyArray_SimpleNew( 1, dmat, NPY_FLOAT64 );

    pi = (int *)a_row->data;
    pj = (int *)a_col->data;
    pv = (double *)a_val->data;

    elem = 0;
    for( i = 0; i < llmat->dim[0]; i++ ) {
      k = llmat->root[i];
      while( k != -1 ) {
        pi[ elem ] = i;
        pj[ elem ] = llmat->col[k];
        pv[ elem ] = llmat->val[k];
        k = llmat->link[k];
        elem++;
      }
    }

    return Py_BuildValue( "OOO",
              PyArray_Return( a_val ),
              PyArray_Return( a_row ),
              PyArray_Return( a_col ) );
}

/* ========================================================================== */

static char CSRMat_To_Coord_Doc[] = "Output coordinates of CSR matrix.";

static PyObject *CSRMat_To_Coord( PyObject *self, PyObject *args ) {

    /* Convert an LL matrix into coordinate format */

    PyObject      *mat;                   /* Input matrix */
    CSRMatObject  *csrmat;                /* Input matrix in CSR format  */
    PyArrayObject *a_row, *a_col, *a_val; /* Matrix in coordinate format */
    int            dmat[1];               /* Dimension descriptor */
    int           *pi, *pj;               /* Intermediate pointers to matrix data */
    double        *pv;
    int            i, j, elem;

    /* Input matrix is the only argument */
    if( !PyArg_ParseTuple( args, "O", &mat ) )
        return NULL;

    csrmat  = (CSRMatObject *)mat;
    dmat[0] = csrmat->nnz;

    /* Allocate numarrays */
    a_row = (PyArrayObject *)PyArray_SimpleNew( 1, dmat, NPY_INT32 );
    a_col = (PyArrayObject *)PyArray_SimpleNew( 1, dmat, NPY_INT32 );
    a_val = (PyArrayObject *)PyArray_SimpleNew( 1, dmat, NPY_FLOAT64 );

    pi = (int *)a_row->data;
    pj = (int *)a_col->data;
    pv = (double *)a_val->data;

    elem = 0;
    for( i = 0; i < csrmat->dim[0]; i++ ) {
        for( j = csrmat->ind[i]; j < csrmat->ind[i+1]; j++ ) {
            pi[ elem ] = i;
            pj[ elem ] = csrmat->col[j];
            pv[ elem ] = csrmat->val[j];
            elem++;
        }
    }

    return Py_BuildValue( "OOO",
              PyArray_Return( a_val ),
              PyArray_Return( a_row ),
              PyArray_Return( a_col ) );
}

/* ========================================================================== */

static char CSRMat_usolve_Doc[] = "Solve upper triangular system U x = b\
\nwhere U is a nonsingular upper triangular csr_mat";

static PyObject* CSRMat_usolve( PyObject *self, PyObject *args ) {

    PyObject      *mat;                                       /* Input matrix */
    CSRMatObject  *csrmat;                      /* Input matrix in CSR format */
    PyArrayObject *a_rhs;                                  /* Right-hand side */
    double        *prhs;                 /* Pointer to right-hand side values */
    PyArrayObject *a_x;                                    /* Solution vector */
    double        *px;                   /* Pointer to solution vector values */
    double         uii;             /* Successive values on the diagonal of U */
    double         uij;             /* Successive values on the i-th row of U */
    int            i, j, k;

    /* Read input matrix, right-hand side and solution vector */
    if( !PyArg_ParseTuple( args, "OO!O!:usolve", &mat,
                                                 &PyArray_Type, &a_rhs,
                                                 &PyArray_Type, &a_x ) )
        return NULL;

    csrmat = (CSRMatObject *)mat;
    prhs   = (double *)a_rhs->data;
    px     = (double *)a_x->data;

    /* Ensure x is initialized to rhs */
    for( i = 0; i < csrmat->dim[0]; i++ ) px[i] = prhs[i];
    
    /* Parse matrix from bottom up */
    for( i = csrmat->dim[0]-1; i > - 1; i-- ) {
        j = csrmat->col[ csrmat->ind[i] ];
        if( j < i ) {
            PyErr_SetString( PyExc_TypeError, "Matrix is not upper triangular" );
            return NULL;
        }
        if( j > i ) {
            PyErr_SetString( PyExc_ZeroDivisionError, "Matrix is singular" );
            return NULL;
        }

        uii = csrmat->val[ csrmat->ind[i] ];

        if( uii == ZERO ) {
            PyErr_SetString( PyExc_ZeroDivisionError, "Matrix is singular" );
            return NULL;
        }
        for( k = csrmat->ind[i]+1; k < csrmat->ind[i+1]; k++ ) {
            j   = csrmat->col[k];
            uij = csrmat->val[k];
            px[i] -= uij * px[j];
        }
        px[i] /= uii;
    }
    
    Py_INCREF( Py_None );
    return Py_None;
}

/* ========================================================================== */

static char CSRMat_utsolve_Doc[] = "Solve lower triangular system U' x = b\
\nwhere U is a nonsingular upper triangular csr_mat";

static PyObject* CSRMat_utsolve( PyObject *self, PyObject *args ) {

    PyObject      *mat;                                       /* Input matrix */
    CSRMatObject  *csrmat;                      /* Input matrix in CSR format */
    PyArrayObject *a_rhs;                                  /* Right-hand side */
    double        *prhs;                 /* Pointer to right-hand side values */
    PyArrayObject *a_x;                                    /* Solution vector */
    double        *px;                   /* Pointer to solution vector values */
    double         uii;             /* Successive values on the diagonal of U */
    double         uij;             /* Successive values on the i-th row of U */
    double         temp;
    int            i, j, k;

    /* Read input matrix, right-hand side and solution vector */
    if( !PyArg_ParseTuple( args, "OO!O!:utsolve", &mat,
                                                  &PyArray_Type, &a_rhs,
                                                  &PyArray_Type, &a_x ) )
        return NULL;

    csrmat = (CSRMatObject *)mat;
    prhs   = (double *)a_rhs->data;
    px     = (double *)a_x->data;

    /* Ensure x is initialized to rhs */
    for( i = 0; i < csrmat->dim[0]; i++ ) px[i] = prhs[i];
    
    /* Parse matrix from top down */
    /* Instead of computing the component of x corresponding to the row
     * being parsed, we update the components of x corresponding to the
     * column indices of the nonzeros in the row being parsed */
    for( i = 0; i < csrmat->dim[0]; i++ ) {
        
        uii = csrmat->val[ csrmat->ind[i] ];
        if( uii == ZERO ) {
            PyErr_SetString( PyExc_ZeroDivisionError, "Matrix is singular" );
            return NULL;
        }
        temp = px[i]/uii;

        for( k = csrmat->ind[i]; k < csrmat->ind[i+1]; k++ ) {
            j   = csrmat->col[k];
            
            if( j < i ) {
                PyErr_SetString( PyExc_TypeError, "Matrix is not upper triangular" );
                return NULL;
            }
            
            uij = csrmat->val[k];
            px[j] -= uij * temp;
        }

        px[i] = temp;
    }
    
    Py_INCREF( Py_None );
    return Py_None;
}

/* ========================================================================== */

static char CSRMat_lsolve_Doc[] = "Solve lower triangular system L x = b\
\nwhere L is a nonsingular lower triangular csr_mat";

static PyObject* CSRMat_lsolve( PyObject *self, PyObject *args ) {

    PyObject      *mat;                                       /* Input matrix */
    CSRMatObject  *csrmat;                      /* Input matrix in CSR format */
    PyArrayObject *a_rhs;                                  /* Right-hand side */
    double        *prhs;                 /* Pointer to right-hand side values */
    PyArrayObject *a_x;                                    /* Solution vector */
    double        *px;                   /* Pointer to solution vector values */
    double         lii;             /* Successive values on the diagonal of L */
    double         lij;             /* Successive values on the i-th row of L */
    int            i, j, k;

    /* Read input matrix, right-hand side and solution vector */
    if( !PyArg_ParseTuple( args, "OO!O!:lsolve", &mat,
                                                 &PyArray_Type, &a_rhs,
                                                 &PyArray_Type, &a_x ) )
        return NULL;

    csrmat = (CSRMatObject *)mat;
    prhs   = (double *)a_rhs->data;
    px     = (double *)a_x->data;

    /* Ensure x is initialized to rhs */
    for( i = 0; i < csrmat->dim[0]; i++ ) px[i] = prhs[i];
    
    /* Parse matrix from top down */
    for( i = 0; i < csrmat->dim[0]; i++ ) {

        for( k = csrmat->ind[i]; k < csrmat->ind[i+1]-1; k++ ) {
            j   = csrmat->col[k];
            
            if( j >= i ) {
                PyErr_SetString( PyExc_TypeError, "Matrix is not lower triangular" );
                return NULL;
            }
            
            lij = csrmat->val[k];
            px[i] -= lij * px[j];
        }

        lii = csrmat->val[ csrmat->ind[i+1]-1 ];
        if( lii == ZERO ) {
            PyErr_SetString( PyExc_ZeroDivisionError, "Matrix is singular" );
            return NULL;
        }

        px[i] /= lii;
    }
    
    Py_INCREF( Py_None );
    return Py_None;
}

/* ========================================================================== */

static char CSRMat_ltsolve_Doc[] = "Solve upper triangular system L' x = b\
\nwhere L is a nonsingular lower triangular csr_mat";

static PyObject* CSRMat_ltsolve( PyObject *self, PyObject *args ) {

    PyObject      *mat;                                       /* Input matrix */
    CSRMatObject  *csrmat;                      /* Input matrix in CSR format */
    PyArrayObject *a_rhs;                                  /* Right-hand side */
    double        *prhs;                 /* Pointer to right-hand side values */
    PyArrayObject *a_x;                                    /* Solution vector */
    double        *px;                   /* Pointer to solution vector values */
    double         lii;             /* Successive values on the diagonal of L */
    double         lij;             /* Successive values on the i-th row of L */
    double         temp;
    int            i, j, k;

    /* Read input matrix, right-hand side and solution vector */
    if( !PyArg_ParseTuple( args, "OO!O!:usolve", &mat,
                                                 &PyArray_Type, &a_rhs,
                                                 &PyArray_Type, &a_x ) )
        return NULL;

    csrmat = (CSRMatObject *)mat;
    prhs   = (double *)a_rhs->data;
    px     = (double *)a_x->data;

    /* Ensure x is initialized to rhs */
    for( i = 0; i < csrmat->dim[0]; i++ ) px[i] = prhs[i];
    
    /* Parse matrix from bottom up */
    for( i = csrmat->dim[0]-1; i > - 1; i-- ) {
        
        j = csrmat->col[ csrmat->ind[i] ];
        
        if( j > i ) {
            PyErr_SetString( PyExc_TypeError, "Matrix is not lower triangular" );
            return NULL;
        }

        j = csrmat->col[ csrmat->ind[i+1]-1 ];
        if( j < i ) {
            PyErr_SetString( PyExc_TypeError, "Matrix is singular" );
            return NULL;
        }

        lii = csrmat->val[ csrmat->ind[i+1]-1 ];
        temp = px[i]/lii;

        if( lii == ZERO ) {
            PyErr_SetString( PyExc_ZeroDivisionError, "Matrix is singular" );
            return NULL;
        }

        for( k = csrmat->ind[i]; k < csrmat->ind[i+1]; k++ ) {
            j   = csrmat->col[k];
            lij = csrmat->val[k];
            px[j] -= lij * temp;
        }
        px[i] = temp;
    }
    
    Py_INCREF( Py_None );
    return Py_None;
}

/* ========================================================================== */

/*
 *      D e f i n i t i o n   o f   S p a r s e T o o l s   m e t h o d s
 */

/* ========================================================================== */

static PyMethodDef SparseToolsMethods[] = {
    { "find_ll",     LLMat_To_Coord,       METH_VARARGS, LLMat_To_Coord_Doc  },
    { "find_csr",    CSRMat_To_Coord,      METH_VARARGS, CSRMat_To_Coord_Doc },
    { "usolve_csr",  CSRMat_usolve,        METH_VARARGS, CSRMat_usolve_Doc   },
    { "utsolve_csr", CSRMat_utsolve,       METH_VARARGS, CSRMat_utsolve_Doc  },
    { "lsolve_csr",  CSRMat_lsolve,        METH_VARARGS, CSRMat_lsolve_Doc   },
    { "ltsolve_csr", CSRMat_ltsolve,       METH_VARARGS, CSRMat_ltsolve_Doc  },
    { NULL,          NULL,                 0,            NULL                }
};

/* ========================================================================== */

void init_sparsetools( void ) {

    (void)Py_InitModule3( "_sparsetools", SparseToolsMethods, "Sparse Matrix Tools" );

    import_array( );         /* Initialize the Numarray module. */
    import_spmatrix( );      /* Initialize the PySparse module. */

    /* Check for errors */
    if (PyErr_Occurred())
    Py_FatalError("Unable to initialize module sparsetools");

    return;
}

/* ========================================================================== */

/*
 *                 E n d   o f   m o d u l e   S p a r s e T o o l s
 */

/* ========================================================================== */
